"""HTTP related utilities."""

from typing import TYPE_CHECKING, Any

import httpx

if TYPE_CHECKING:
    from .types import URLTypes

HTTPStatusError = httpx.HTTPStatusError
HTTPNetworkError = httpx.NetworkError
HTTPTimeoutException = httpx.TimeoutException
HTTPResponse = httpx.Response
HTTPRequest = httpx.Request


class Method:
    """HTTP methods, trying to prevent mistakes with this."""

    PROPFIND = "PROPFIND"
    PROPPATCH = "PROPPATCH"
    MKCOL = "MKCOL"
    COPY = "COPY"
    MOVE = "MOVE"
    LOCK = "LOCK"
    UNLOCK = "UNLOCK"
    DELETE = "DELETE"
    GET = "GET"
    PUT = "PUT"


def request(method: str):  # type: ignore[no-untyped-def]
    """Extending with new verb `method`."""

    def func(client: "Client", url: "URLTypes", **kwargs: Any) -> "HTTPResponse":
        return client.request(method, url, **kwargs)

    return func


class Client(httpx.Client):
    """HTTP client with additional verbs for the Webdav."""

    propfind = request(Method.PROPFIND)
    proppatch = request(Method.PROPPATCH)
    mkcol = request(Method.MKCOL)
    copy = request(Method.COPY)
    move = request(Method.MOVE)
    lock = request(Method.LOCK)
    unlock = request(Method.UNLOCK)


# Non-standard header used in Apache Web Server/cPanel
# often used by shared hosting providers to limit the bandwidth of customers
BANDWIDTH_LIMIT_EXCEEDED = 509, "Bandwidth Limit Exceeded"



"""Date parsing utilities."""

from email.utils import parsedate_to_datetime
from typing import TYPE_CHECKING

from dateutil.parser import parse

if TYPE_CHECKING:
    from datetime import datetime


def fromisoformat(datetime_string: str) -> "datetime":
    """Convert ISO 8601 datetime string to datetime object."""
    return parse(datetime_string)


def from_rfc1123(datetime_string: str) -> "datetime":
    """Convert rfc1123 datetime string to datetime object."""
    try:
        return parsedate_to_datetime(datetime_string)
    except Exception:  # noqa: BLE001
        # fallback in case ^ is unable to parse the datetime string
        return parse(datetime_string)



"""Client for the webdav."""

import locale
import shutil
import threading
from contextlib import contextmanager, suppress
from http import HTTPStatus
from io import TextIOWrapper
from typing import (
    TYPE_CHECKING,
    Any,
    BinaryIO,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Set,
    TextIO,
    Union,
    cast,
)

from .callback import wrap_file_like
from .fs_utils import peek_filelike_length
from .func_utils import wrap_fn
from .http import Client as HTTPClient
from .http import HTTPStatusError
from .http import Method as HTTPMethod
from .multistatus import (
    MultiStatusResponseError,
    Response,
    parse_multistatus_response,
    prepare_propfind_request_data,
)
from .retry import retry as _retry
from .stream import IterStream, read_chunks
from .urls import URL, join_url

if TYPE_CHECKING:
    from datetime import datetime
    from os import PathLike
    from typing import AnyStr

    from .multistatus import DAVProperties, MultiStatusResponse
    from .retry import RetryFunc
    from .types import AuthTypes, HeaderTypes, HTTPResponse, URLTypes


DEFAULT_CHUNK_SIZE = 2**22


def _prepare_result_info(
    response: Response, base_url: URL, detail: bool = True
) -> Union[str, Dict[str, Any]]:
    """Transform response to a dictionary/str for info/ls."""
    rel = response.path_relative_to(base_url)
    if not detail:
        return rel
    return {
        "name": rel,
        "href": response.href,
        **response.properties.as_dict(),
    }


class ClientError(Exception):
    """Custom exception thrown by the Client."""

    def __init__(self, msg: str) -> None:
        """Instantiate exception with a msg."""
        self.msg: str = msg
        super().__init__(msg)

    def __str__(self) -> str:
        """Provide str repr of the msg."""
        return self.msg


class ResourceConflict(ClientError):  # noqa: N818
    """Raised when there was conflict during the operation (got 409)."""


class ForbiddenOperation(ClientError):  # noqa: N818
    """Raised when the operation was forbidden (got 403)."""


class ResourceAlreadyExists(ClientError):  # noqa: N818
    """Error returned if the resource already exists."""

    def __init__(self, path: str) -> None:
        """Instantiate exception with the path that already exists."""
        self.path = path
        super().__init__(f"The resource {path} already exists")


class InsufficientStorage(ClientError):  # noqa: N818
    """Error when the resource does not exist on the server."""

    def __init__(self, path: str) -> None:
        """Instantiate exception with the path for which the request failed."""
        self.path = path
        super().__init__("Insufficient Storage on the server")


class BadGatewayError(ClientError):
    """Error when bad gateway error is thrown."""

    def __init__(self) -> None:
        """Raised when 502 status code is raised by the server."""
        msg = "The destination server may have refused to accept the resource"
        super().__init__(msg)


class ResourceLocked(ClientError):  # noqa: N818
    """Error raised when the resource is locked."""


class ResourceNotFound(ClientError):  # noqa: N818
    """Error when the resource does not exist on the server."""

    def __init__(self, path: str) -> None:
        """Instantiate exception with path that does not exist."""
        self.path = path
        super().__init__(f"The resource {path} could not be found in the server")


class HTTPError(ClientError):
    """Custom Exception for our HTTPStatusError."""

    def __init__(self, response: "HTTPResponse") -> None:
        """Instantiate exception with the failed response."""
        self.response = response
        self.status_code = response.status_code
        self.request = response.request

        super().__init__(f"received {self.status_code} ({self.response.reason_phrase})")


class IsAResourceError(ClientError):
    """Exception thrown when the path is a resource.

    Could be thrown when the collection is expected.
    """

    def __init__(self, path: str, msg: str = "") -> None:
        """Initialize with the path and the appropriate message."""
        self.path: str = path
        super().__init__(f"{path} is a collection. {msg}")


class IsACollectionError(ClientError):
    """Exception thrown when the path is a collection.

    Could be thrown when the resource/non-collection is expected.
    """

    def __init__(self, path: str, msg: str = "") -> None:
        """Initialize with the path and the appropriate message."""
        self.path: str = path
        super().__init__(f"{path} is a collection. {msg}")


class MultiStatusError(ClientError):
    """Wrapping MultiStatusResponseError with ClientError."""


class FeatureDetection:
    """Detect features in the webdav resources.

    Mostly used for detecting support for Accept-Ranges as ownCloud/NextCloud
    don't advertise support for it in GET requests.
    """

    supports_ranges: bool
    dav_compliances: Set[str]

    def __init__(self, options_response: Optional["HTTPResponse"] = None) -> None:
        """Initialize with the given response."""
        dav_compliances = set()
        supports_ranges = False
        if options_response:
            dav_header = options_response.headers.get("dav", "")
            dav_compliances = {f.strip() for f in dav_header.split(",")}
            supports_ranges = options_response.headers.get("accept-ranges") == "bytes"

        self.dav_compliances = dav_compliances
        self.supports_ranges = supports_ranges


class Client:
    """Provides higher level APIs for interacting with Webdav server."""

    def __init__(
        self,
        base_url: "URLTypes",
        auth: Optional["AuthTypes"] = None,
        http_client: Optional["HTTPClient"] = None,
        retry: Union["RetryFunc", bool] = True,
        chunk_size: int = DEFAULT_CHUNK_SIZE,
        **client_opts: Any,
    ) -> None:
        """Instantiate client for webdav.

        Examples:
            >>> client = Client("https://webdav.example.org")
            >>> client.ls("/")

        Args:
            base_url: base url of the Webdav server
            auth: Auth for the webdav.
                Auth can be any of the following:

                - a tuple of (user, password)
                - None if no auth is required.

                Refer to `Customizing Authentication \
                <https://www.python-httpx.org/advanced/ \
                #customizing-authentication>`_
                for more options.

            http_client: http client to use instead, useful in mocking
                (when extending, it is expected to have implemented additional
                verbs from webdav)
            retry: disable or enable retry on client. Can also pass a callable
                to handle it there. Some well-known errors are handled and
                retried a few times with the backoff.

        All of the following keyword arguments are passed along to the
        `httpx <https://www.python-httpx.org/api/#client>`_, the http library
        this client is built on.

        Keyword Args:
            headers: Dict. of HTTP headers to include when sending requests
            cookies: Dict. of Cookie items to include when sending requests
            verify: SSL certificates used to verify the identity of requested
                hosts. Can be any of:

                - True (uses default CA bundle),
                - a path to an SSL certificate file,
                - False (disable verification), or
                - a :py:class:`ssl.SSLContext`

            cert: An SSL certificate used by the requested host to
                authenticate the client.
                Either a path to an SSL certificate file,
                or two-tuple of (certificate file, key file),
                or a three-tuple of (certificate file, key file, password).
            proxies: A dictionary mapping proxy keys to proxy URLs
            timeout: The timeout configuration to use when sending requests
            limits: The limits configuration to use
            max_redirects: The maximum number of redirect responses that
                should be followed
            trust_env: Enables or disables usage of environment variables
                for configuration
        """
        client_opts.update({"base_url": base_url, "auth": auth})
        self.http: HTTPClient = http_client or HTTPClient(**client_opts)
        self.base_url = URL(base_url)
        self.with_retry = retry if callable(retry) else _retry(retry)
        self._detected_features: Optional[FeatureDetection] = None
        self._detect_feature_lock = threading.RLock()
        self.chunk_size = chunk_size

    @property
    def detected_features(self) -> FeatureDetection:
        """Feature detection for the server."""
        if not self._detected_features:
            with self._detect_feature_lock:
                # a lot of threads might be stuck on thread lock
                # and if one is done with it, it means we already
                # have it set, so we should look for it rather than
                # sending out a request.
                if self._detected_features:  # pragma: no cover
                    return self._detected_features

                resp = None
                with suppress(Exception):
                    resp = self.http.options(self.base_url)
                self._detected_features = FeatureDetection(resp)
        return self._detected_features

    def options(self, path: str = "") -> Set[str]:
        """Returns features detected in the webdav server."""
        resp = self.http.options(path)
        detected_features = FeatureDetection(resp)
        return detected_features.dav_compliances

    def join_url(self, path: str, add_trailing_slash: bool = False) -> URL:
        """Join resource path with base url of the webdav server."""
        return join_url(self.base_url, path, add_trailing_slash=add_trailing_slash)

    def propfind(
        self,
        path: str,
        data: Optional[str] = None,
        headers: Optional["HeaderTypes"] = None,
        follow_redirects: bool = False,
    ) -> "MultiStatusResponse":
        """Returns properties of the specific resource by propfind request."""
        call = wrap_fn(
            self._request,
            HTTPMethod.PROPFIND,
            path,
            content=data,
            headers=headers,
            follow_redirects=follow_redirects,
        )
        http_resp = self.with_retry(call)
        return parse_multistatus_response(http_resp)

    def get_props(
        self,
        path: str,
        name: Optional[str] = None,
        namespace: Optional[str] = None,
        data: Optional[str] = None,
    ) -> "DAVProperties":
        """Returns properties of a resource by doing a propfind request.

        Can also selectively request the properties by passing name or data.
        """
        data = data or prepare_propfind_request_data(name, namespace)
        headers = {"Content-Type": "application/xml"} if data else {}
        result = self.propfind(path, headers=headers, data=data)
        response = result.get_response_for_path(self.base_url.path, path)
        return response.properties

    def get_property(
        self, path: str, name: str, namespace: Optional[str] = None
    ) -> Any:
        """Returns appropriate property from the propfind response.

        Also supports getting named properties
        (for now restricted to a single string with the given namespace)
        """
        props = self.get_props(path, name=name, namespace=namespace)
        return getattr(props, name, "")

    def set_property(self) -> None:
        """Setting additional property to a resource."""

    def _request(
        self,
        method: str,
        path: str,
        add_trailing_slash: bool = False,
        **kwargs: Any,
    ) -> "HTTPResponse":
        """Internal method for sending request to the server.

        It handles joining path correctly and checks for common http errors.
        """
        url = self.join_url(path, add_trailing_slash=add_trailing_slash)
        http_resp = self.http.request(method, url, **kwargs)

        if http_resp.status_code == HTTPStatus.NOT_FOUND:
            raise ResourceNotFound(path)
        if http_resp.status_code == HTTPStatus.INSUFFICIENT_STORAGE:
            raise InsufficientStorage(path)
        if http_resp.status_code == HTTPStatus.BAD_GATEWAY:
            raise BadGatewayError

        try:
            http_resp.raise_for_status()
        except HTTPStatusError as exc:
            raise HTTPError(http_resp) from exc

        return http_resp

    def request(self, method: str, path: str, **kwargs: Any) -> "HTTPResponse":
        """Sends request to a server with given method and path.

        Also checks for Multistatus response and other http errors.
        """
        http_resp = self._request(method, path, **kwargs)
        if http_resp.status_code == HTTPStatus.MULTI_STATUS:
            # if it's 207, it's most likely an error
            # or, a partial success (however you see it).
            # except for the propfind, for which we use `_request` directly)
            # in the above `propfind` function.
            result = parse_multistatus_response(http_resp)
            try:
                result.raise_for_status()
            except MultiStatusResponseError as exc:
                raise MultiStatusError(exc.msg) from exc

        return http_resp

    def move(self, from_path: str, to_path: str, overwrite: bool = False) -> None:
        """Move resource to a new destination (with or without overwriting)."""
        return self._transfer(HTTPMethod.MOVE, from_path, to_path, overwrite=overwrite)

    def _transfer(
        self,
        operation: str,
        from_path: str,
        to_path: str,
        overwrite: bool,
        depth: Union[int, str] = "infinity",
    ) -> None:
        """Transfer a resource by copying/moving from a path to the other."""
        assert operation in {HTTPMethod.MOVE, HTTPMethod.COPY}

        to_url = self.join_url(to_path)
        headers = {
            "Destination": str(to_url),
            "Overwrite": "T" if overwrite else "F",
            "Depth": str(depth),
        }

        call = wrap_fn(self.request, operation, from_path, headers=headers)
        try:
            self.with_retry(call)
        except HTTPError as exc:
            if exc.status_code == HTTPStatus.FORBIDDEN:
                msg = "the source and the destination could be the same"
                raise ForbiddenOperation(msg) from exc
            if exc.status_code == HTTPStatus.CONFLICT:
                msg = (
                    "there was a conflict when trying to "
                    f"{operation.lower()} the resource"
                )
                raise ResourceConflict(msg) from exc
            if exc.status_code == HTTPStatus.PRECONDITION_FAILED:
                raise ResourceAlreadyExists(to_path) from exc
            if exc.status_code == HTTPStatus.LOCKED:
                msg = "the source or the destination resource is locked"
                raise ResourceLocked(msg) from exc

            raise

    def copy(
        self,
        from_path: str,
        to_path: str,
        depth: Union[int, str] = "infinity",
        overwrite: bool = False,
    ) -> None:
        """Copy resource."""
        return self._transfer(
            HTTPMethod.COPY,
            from_path,
            to_path,
            depth=depth,
            overwrite=overwrite,
        )

    def mkdir(self, path: str) -> None:
        """Create a collection."""
        call = wrap_fn(self.request, HTTPMethod.MKCOL, path, add_trailing_slash=True)
        try:
            http_resp = self.with_retry(call)
        except HTTPError as exc:
            if exc.status_code == HTTPStatus.METHOD_NOT_ALLOWED:
                raise ResourceAlreadyExists(path) from exc
            if exc.status_code == HTTPStatus.FORBIDDEN:
                msg = (
                    "the server does not allow creation in the namespace "
                    "or cannot accept members"
                )
                raise ForbiddenOperation(msg) from exc
            if exc.status_code == HTTPStatus.CONFLICT:
                msg = "parent of the collection does not exist"
                raise ResourceConflict(msg) from exc

            raise

        assert http_resp.status_code in (HTTPStatus.OK, HTTPStatus.CREATED)

    def remove(self, path: str) -> None:
        """Remove a resource."""
        call = wrap_fn(self.request, HTTPMethod.DELETE, path)
        try:
            self.with_retry(call)
        except HTTPError as exc:
            if exc.status_code == HTTPStatus.LOCKED:
                raise ResourceLocked("the resource is locked") from exc
            raise

    def ls(
        self,
        path: str,
        detail: bool = True,
        allow_listing_resource: bool = True,
    ) -> List[Union[str, Dict[str, Any]]]:
        """List items in a resource/collection.

        Args:
            path: Path to the resource
            detail: If detail=True, additional information is returned
                in a dictionary
            allow_listing_resource: If True and path is a resource
                (non-collection), ls will return the file entry/details.
                Otherwise, it will raise an error.
        """
        result = self.propfind(path, headers={"Depth": "1"}, follow_redirects=True)
        responses = result.responses

        url = self.join_url(path)
        response = responses.get(url.path)

        if response:
            typ = response.properties.resource_type
            if typ == "file" and not allow_listing_resource:
                raise IsAResourceError(path, "cannot list from a resource itself")
            if typ == "directory":
                responses.pop(url.path)
        else:  # pragma: no cover
            assert not response  # response should always be there

        return [
            _prepare_result_info(resp, self.base_url, detail)
            for resp in responses.values()
        ]

    def info(self, path: str) -> Dict[str, Any]:
        """Returns information about the path itself."""
        result = self.propfind(path, headers={"Depth": "1"})
        responses = result.responses

        url = self.join_url(path)
        details = _prepare_result_info(responses[url.path], self.base_url, detail=True)
        assert not isinstance(details, str)
        return details

    def exists(self, path: str) -> bool:
        """Checks whether the resource with the given path exists or not."""
        try:
            self.propfind(path)
        except ResourceNotFound:
            return False

        return True

    def isdir(self, path: str) -> bool:
        """Checks whether the resource with the given path is a directory."""
        return bool(self.get_props(path).collection)

    def isfile(self, path: str) -> bool:
        """Checks whether the resource with the given path is a file."""
        return not self.isdir(path)

    def content_length(self, path: str) -> Optional[int]:
        """Returns content-length of the resource with the given path."""
        return self.get_props(path, "content_length").content_length

    def created(self, path: str) -> Optional["datetime"]:
        """Returns creationdate of the resource with the given path."""
        return self.get_props(path, "created").created

    def modified(self, path: str) -> Optional["datetime"]:
        """Returns getlastmodified of the resource with the given path."""
        return self.get_props(path, "modified").modified

    def etag(self, path: str) -> Optional[str]:
        """Returns etag of the resource with the given path."""
        return self.get_props(path, "etag").etag

    def content_type(self, path: str) -> Optional[str]:
        """Returns content type of the resource with the given path."""
        return self.get_props(path, "content_type").content_type

    def content_language(self, path: str) -> Optional[str]:
        """Returns content language of the resource with the given path."""
        return self.get_props(path, "content_language").content_language

    @contextmanager
    def open(
        self,
        path: str,
        mode: str = "r",
        encoding: Optional[str] = None,
        chunk_size: Optional[int] = None,
    ) -> Iterator[Union[TextIO, BinaryIO]]:
        """Returns file-like object to a resource."""
        if self.isdir(path):
            raise IsACollectionError(path, "Cannot open a collection")
        assert mode in {"r", "rt", "rb"}

        with IterStream(
            self,
            self.join_url(path),
            chunk_size=chunk_size or self.chunk_size,
        ) as buffer:
            buff = cast("BinaryIO", buffer)

            if mode == "rb":
                yield buff
            else:
                encoding = (
                    encoding or buffer.encoding or locale.getpreferredencoding(False)
                )
                yield TextIOWrapper(buff, encoding=encoding)

    def download_fileobj(
        self,
        from_path: str,
        file_obj: BinaryIO,
        callback: Optional[Callable[[int], Any]] = None,
        chunk_size: Optional[int] = None,
    ) -> None:
        """Write stream from path to given file object."""
        with self.open(from_path, mode="rb", chunk_size=chunk_size) as remote_obj:
            wrapped = wrap_file_like(file_obj, callback, method="write")
            shutil.copyfileobj(remote_obj, wrapped)  # type: ignore[misc]

    def download_file(
        self,
        from_path: str,
        to_path: "PathLike[AnyStr]",
        chunk_size: Optional[int] = None,
        callback: Optional[Callable[[int], Any]] = None,
    ) -> None:
        """Download file from remote path to local path."""
        with open(to_path, mode="wb") as fobj:
            self.download_fileobj(
                from_path, fobj, callback=callback, chunk_size=chunk_size
            )

    def upload_file(  # noqa: PLR0913
        self,
        from_path: "PathLike[AnyStr]",
        to_path: str,
        overwrite: bool = False,
        chunk_size: Optional[int] = None,
        callback: Optional[Callable[[int], Any]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> None:
        """Upload file from local path to a given remote path."""
        with open(from_path, mode="rb") as fobj:
            self.upload_fileobj(
                fobj,
                to_path,
                overwrite=overwrite,
                chunk_size=chunk_size,
                callback=callback,
                headers=headers,
            )

    def upload_fileobj(  # noqa: PLR0913
        self,
        file_obj: BinaryIO,
        to_path: str,
        overwrite: bool = False,
        callback: Optional[Callable[[int], Any]] = None,
        chunk_size: Optional[int] = None,
        size: Optional[int] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> None:
        """Upload file from file object to given path."""
        if headers is None:
            headers = {}

        # we try to avoid chunked transfer as much as possible
        # so we try to use size as a hint if provided.
        # else, we will try to find that out from the file object
        # if we are not successful in that, we gracefully fallback
        # to the chunked encoding.
        if size is None:
            size = peek_filelike_length(file_obj)

        if size is not None:
            headers = {"Content-Length": str(size), **headers}

        if not overwrite and self.exists(to_path):
            raise ResourceAlreadyExists(to_path)

        wrapped = wrap_file_like(file_obj, callback)
        content = read_chunks(wrapped, chunk_size=chunk_size or self.chunk_size)

        http_resp = self.request(
            HTTPMethod.PUT, to_path, content=content, headers=headers
        )
        http_resp.raise_for_status()



"""Retry mechanism for the webdav Client."""

from http import HTTPStatus
from typing import Callable, Protocol, TypeVar, cast

from .func_utils import retry as retry_func
from .http import BANDWIDTH_LIMIT_EXCEEDED

_T = TypeVar("_T")

BACKOFF: float = 1


def filter_errors(exc: Exception) -> bool:
    """Filter these errors and retry if they fall in these categories."""
    from .client import HTTPError  # noqa: PLC0415

    if isinstance(exc, HTTPError):
        return exc.status_code in (
            HTTPStatus.TOO_MANY_REQUESTS,
            HTTPStatus.INTERNAL_SERVER_ERROR,
            HTTPStatus.SERVICE_UNAVAILABLE,
            HTTPStatus.GATEWAY_TIMEOUT,
            BANDWIDTH_LIMIT_EXCEEDED,
        )
    return True


def _exp_backoff(attempt: int) -> float:
    """Backoff exponentially."""
    # for some reason, mypy is unable to figure out types
    return cast("float", BACKOFF * 2**attempt)


class RetryFunc(Protocol):
    """Retry function protocol."""

    def __call__(self, f: Callable[[], _T], /) -> _T:
        """Callable wrapper."""


def retry(arg: bool = False, tries: int = 3) -> RetryFunc:
    """Retry if arg up to `tries` times."""
    from .client import BadGatewayError, HTTPError, ResourceLocked  # noqa: PLC0415

    return retry_func(
        tries if arg is True else 1,
        (ResourceLocked, BadGatewayError, HTTPError),
        timeout=_exp_backoff,
        filter_errors=filter_errors,
    )



"""Handle streaming response for file."""

from contextlib import contextmanager
from functools import partial
from http import HTTPStatus
from io import DEFAULT_BUFFER_SIZE, RawIOBase
from itertools import takewhile
from typing import (
    IO,
    TYPE_CHECKING,
    AnyStr,
    Generator,
    Iterator,
    List,
    Optional,
    Tuple,
    cast,
)

from .func_utils import repeat_func
from .http import HTTPNetworkError, HTTPTimeoutException
from .http import Method as HTTPMethod

if TYPE_CHECKING:
    from typing_extensions import Buffer, Self

    from .client import Client
    from .http import Client as HTTPClient
    from .types import HTTPResponse, URLTypes


def request(client: "HTTPClient", url: "URLTypes", pos: int = 0) -> "HTTPResponse":
    """Streams a file from url from given position."""
    headers = {}
    if pos:
        headers.update({"Range": f"bytes={pos}-"})

    req = client.build_request(HTTPMethod.GET, url, headers=headers)
    return client.send(req, stream=True, follow_redirects=True)


@contextmanager
def iter_url(
    client: "Client",
    url: "URLTypes",
    chunk_size: Optional[int] = None,
    pos: int = 0,
) -> Iterator[Tuple["HTTPResponse", Iterator[bytes]]]:
    """Iterate over chunks requested from url.

    Reopens connection on network failure.
    """

    def gen(
        response: "HTTPResponse",
    ) -> Generator[bytes, None, None]:
        nonlocal pos
        try:
            while True:
                if response.status_code == HTTPStatus.REQUESTED_RANGE_NOT_SATISFIABLE:
                    return  # range request outside file
                response.raise_for_status()

                try:
                    for chunk in response.iter_bytes(chunk_size=chunk_size):
                        pos += len(chunk)
                        yield chunk
                    break
                except (HTTPTimeoutException, HTTPNetworkError):
                    response.close()
                    if not (
                        response.headers.get("Accept-Ranges") == "bytes"
                        or client.detected_features.supports_ranges
                    ):
                        raise
                    response = request(client.http, url, pos=pos)
        finally:
            response.close()

    response = request(client.http, url, pos=pos)
    chunks = gen(response)
    try:
        yield response, chunks
    finally:
        # Ensure connection is closed
        chunks.close()
        response.close()


class IterStream(RawIOBase):
    """Create a streaming file-like object."""

    def __init__(
        self,
        client: "Client",
        url: "URLTypes",
        chunk_size: Optional[int] = None,
    ) -> None:
        """Pass a iterator to stream through."""
        super().__init__()

        self.buffer = b""
        # setting chunk_size is not possible yet with httpx
        # though it is to be released in a new version.
        self.chunk_size = chunk_size or client.chunk_size
        self.client = client
        self.url = url
        self._loc: int = 0
        self._cm = iter_url(client, self.url, chunk_size=chunk_size)
        self._iterator: Optional[Iterator[bytes]] = None
        self._initial_response: Optional[HTTPResponse] = None

    @property
    def supports_ranges(self) -> bool:
        """Checks whether the server supports ranges for the resource.

        Even if it does not advertise, we'll use base url's OPTIONS request
        to see if the server supports Range header or not. And, we want to
        avoid checking that as much as possible.
        """
        response = self._initial_response
        if response and response.headers.get("Accept-Ranges") == "bytes":
            return True
        # consider if checking Accept-Ranges from OPTIONS request on self.url
        # would be a better solution than using base url.
        return self.client.detected_features.supports_ranges

    @property
    def size(self) -> Optional[int]:
        """Size of the file object."""
        assert self._initial_response
        content_length: str = self._initial_response.headers.get("Content-Length", "")
        return int(content_length) if content_length.isdigit() else None

    @property
    def loc(self) -> int:
        """Keep track of location of the stream/file for callbacks."""
        return self._loc

    @loc.setter
    def loc(self, value: int) -> None:
        """Update location, and run callbacks."""
        self._loc = value

    def __enter__(self) -> "Self":
        """Send a streaming response."""
        self._initial_response, self._iterator = self._cm.__enter__()
        return self

    def __exit__(self, *args: object) -> None:
        """Close the response."""
        self.close()

    @property
    def encoding(self) -> Optional[str]:
        """Encoding of the response."""
        assert self._initial_response
        return self._initial_response.encoding

    def close(self) -> None:
        """Close response if not already."""
        if self._iterator:
            self._cm.__exit__(None, None, None)

        self._iterator = None
        self.buffer = b""

    def seek(self, offset: int, whence: int = 0) -> int:
        """Seek the file object."""
        if whence == 0:
            loc = offset
        elif whence == 1:
            if offset >= 0:
                self.read(offset)
                return self.loc
            loc = self.loc + offset
        elif whence == 2:
            if not self.size:
                raise ValueError("cannot seek to the end of file")
            loc = self.size + offset
        else:
            raise ValueError(f"invalid whence ({whence}, should be 0, 1 or 2)")
        if loc < 0:
            raise ValueError("Seek before start of file")
        if loc and not self.supports_ranges:
            raise ValueError("server does not support ranges")

        self.close()
        self._cm = iter_url(self.client, self.url, pos=loc, chunk_size=self.chunk_size)
        _, self._iterator = self._cm.__enter__()
        self.loc = loc
        return loc

    def tell(self) -> int:
        """Return current position of the fileobj."""
        return self.loc

    @property
    def closed(self) -> bool:
        """Check whether the stream was closed or not."""
        return self._iterator is None

    def readable(self) -> bool:
        """Stream is readable."""
        return True

    def seekable(self) -> bool:
        """Stream is not seekable."""
        return True

    def writable(self) -> bool:
        """Stream not writable."""
        return False

    def readall(self) -> bytes:
        """Read all of the bytes."""
        return b"".join(iter(partial(self.read1, -1), b""))

    def read(self, num: int = -1) -> bytes:
        """Read n bytes at max."""
        if num < 0:
            return self.readall()

        buff = b""
        while len(buff) < num:
            chunk = self.read1(num - len(buff))
            if not chunk:
                break
            buff += chunk
        return buff

    def read1(self, num: int = -1) -> bytes:
        """Read at maximum once."""
        assert self._iterator
        try:
            chunk = self.buffer or next(self._iterator)
        except StopIteration:
            return b""

        if num <= 0:
            output, self.buffer = chunk, b""
        else:
            output, self.buffer = chunk[:num], chunk[num:]

        self.loc += len(output)
        return output

    def readinto(self, sequence: "Buffer") -> int:
        """Read into the buffer."""
        return read_into(self, sequence)  # type: ignore

    def readinto1(self, sequence: "Buffer") -> int:
        """Read into the buffer with 1 read at max."""
        return read_into(self, sequence, read_once=True)  # type: ignore


def read_chunks(obj: IO[AnyStr], chunk_size: Optional[int] = None) -> Iterator[AnyStr]:
    """Read file object in chunks."""
    func = partial(obj.read, chunk_size or DEFAULT_BUFFER_SIZE)
    return takewhile(bool, repeat_func(func))


def split_chunk(part: AnyStr, char: AnyStr) -> Tuple[AnyStr, Optional[AnyStr]]:
    """Split chunk into two parts, based on given delimiter character.

    It will return tuple of those two chunks. If no such delim exists,
    the first element of tuple will have the complete chunk whereas the other
    will be None.
    """
    found = part.find(char)
    if found > -1:
        return part[: found + len(char)], part[found + len(char) :]

    return part, None


def read_until(obj: IO[AnyStr], char: str) -> Iterator[AnyStr]:
    """Read chunks until the char is reached."""
    is_bytes = isinstance(obj.read(0), bytes)
    # `c` and the `joiner` should be the same type as the file `obj` is of.
    assert char
    _char: AnyStr = cast("AnyStr", char.encode() if is_bytes else char)
    joiner: AnyStr = cast("AnyStr", b"" if is_bytes else "")

    reader = read_chunks(obj)
    leftover = None
    while True:
        out: List[AnyStr] = []
        while True:
            try:
                part = leftover or next(reader)
            except StopIteration:
                part = joiner

            if not part:
                if out:
                    yield joiner.join(out)
                return

            output, leftover = split_chunk(part, _char)
            out.append(output)
            if leftover is not None:
                break

        yield joiner.join(out)


def read_into(obj: IO[AnyStr], sequence: "Buffer", read_once: bool = False) -> int:
    """Read into the buffer."""
    out = memoryview(sequence).cast("B")
    func = obj.read1 if read_once else obj.read  # type: ignore
    data = func(out.nbytes)

    out[: len(data)] = data  # type: ignore[assignment]
    return len(data)



"""Parsing propfind response."""

from http.client import responses
from typing import TYPE_CHECKING, Any, Dict, Optional, Union
from xml.etree.ElementTree import Element, ElementTree, SubElement
from xml.etree.ElementTree import fromstring as str2xml
from xml.etree.ElementTree import tostring as xml2string

from .date_utils import from_rfc1123, fromisoformat
from .urls import URL, join_url_path, relative_url_to, strip_trailing_slash

if TYPE_CHECKING:
    from httpx import Response as HTTPResponse


# Map name used in library with actual prop name
MAPPING_PROPS: Dict[str, str] = {
    "content_length": "getcontentlength",
    "etag": "getetag",
    "created": "creationdate",
    "modified": "getlastmodified",
    "content_language": "getcontentlanguage",
    "content_type": "getcontenttype",
    "display_name": "displayname",
}


def prop(
    node: Union[Element, ElementTree], name: str, relative: bool = False
) -> Optional[str]:
    """Returns text of the property if it exists under DAV namespace."""
    namespace = "{DAV:}"
    selector = ".//" if relative else ""
    xpath = f"{selector}{namespace}{name}"
    return node.findtext(xpath)


class DAVProperties:
    """Parses <d:propstat> data into certain properties.

    Only supports a certain set of properties to extract. Others are ignored.
    """

    def __init__(self, response_xml: Optional[Element] = None):
        """Parses props to certain attributes.

        Args:
             response_xml: <d:propstat> element
        """
        self.response_xml: Optional[Element] = response_xml
        self.raw: Dict[str, Any] = {}

        def extract_text(prop_name: str) -> Optional[str]:
            text = (
                prop(response_xml, MAPPING_PROPS[prop_name], relative=True)
                if response_xml is not None
                else None
            )
            self.raw[prop_name] = text
            return text

        created = extract_text("created")
        self.created = fromisoformat(created) if created else None

        modified = extract_text("modified")
        self.modified = from_rfc1123(modified) if modified else None

        self.etag = extract_text("etag") or None
        self.content_type = extract_text("content_type")

        content_length = extract_text("content_length")
        self.content_length = int(content_length) if content_length else None
        self.content_language = extract_text("content_language")

        collection: Optional[bool] = None
        resource_type: Optional[str] = None
        resource_xml: Optional[Element] = None

        if response_xml is not None:
            resource_xml = response_xml.find(".//{DAV:}resourcetype")

        if resource_xml is not None:
            collection = resource_xml.find(".//{DAV:}collection") is not None
            resource_type = "directory" if collection else "file"

        self.collection = collection
        self.resource_type = resource_type

        self.display_name = extract_text("display_name")

    def as_dict(self, raw: bool = False) -> Dict[str, Any]:
        """Returns all properties that it supports parsing.

        Args:
            raw: Provides raw data instead.
        """
        if raw:
            return self.raw

        return {
            "content_length": self.content_length,
            "created": self.created,
            "modified": self.modified,
            "content_language": self.content_language,
            "content_type": self.content_type,
            "etag": self.etag,
            "type": self.resource_type,
            "display_name": self.display_name,
        }


class Response:
    """Individual response from multistatus propfind response."""

    def __init__(self, response_xml: Element) -> None:
        """Parses xml from each responses to an easier format.

        Args:
            response_xml: <d:response> element

        Note: we do parse <d:propstat> to figure out status,
        but we leave <d:prop> to ResourceProps to figure out.
        """
        self.response_xml = response_xml
        href = prop(response_xml, "href")
        assert href

        parsed = URL(href)

        # href is what was received
        self.href = href
        self.is_href_absolute = parsed.is_absolute_url

        # path is absolute path from the href
        # collection might contain `/` at the end
        self.path = parsed.path

        # path_norm is the path-absolute without differentiating
        # `/` at the end. Used as a key for the responses.
        # but does have a slash in front.
        self.path_norm = strip_trailing_slash(self.path)

        status_line = prop(response_xml, "status")
        code = None
        if status_line:
            _, code_str, *_ = status_line.split()
            code = int(code_str)

        self.status_code = code
        self.reason_phrase = responses[self.status_code] if self.status_code else None

        self.response_description: Optional[str] = prop(
            response_xml, "responsedescription"
        )
        self.error = prop(response_xml, "error")
        self.location = prop(response_xml, "location")

        self.has_propstat = response_xml.find("{DAV:}propstat") is not None
        self.properties = DAVProperties(response_xml)

    def __str__(self) -> str:
        """User representation for the response."""
        return f"Response: {self.path_norm}"

    def __repr__(self) -> str:
        """Repr for the response."""
        return f"Response: {self.path}"

    def path_relative_to(self, base_url: URL) -> str:
        """Relative path of the resource from the response."""
        return relative_url_to(base_url, self.path_norm)


class MultiStatusResponseError(Exception):
    """Raised when multistatus response has failures in it."""

    def __init__(self, statuses: Dict[str, str]) -> None:
        """Pass multiple statuses, which is displayed when error is raised."""
        self.statuses = statuses

        assert self.statuses
        if len(self.statuses) > 1:
            msg = "multiple errors received: " + str(statuses)
        else:
            path, status = next(iter(statuses.items()))
            msg = f"The resource {path} is {status.lower()}"

        self.msg = msg
        super().__init__(msg)


class MultiStatusResponse:
    """Parse multistatus responses from the received http content.

    Propfind response can contain multiple responses for multiple resources.
    The format is in xml, so we try to parse it into an easier format, and
    provide an easier way to access a response for one particular resource.

    Also note that propfind response could be partial, in that those
    properties may not exist if we are doing propfind with named properties.
    """

    def __init__(self, content: str) -> None:
        """Parse the http content from propfind request.

        Args:
             content: body received from PROPFIND call
        """
        self.content = content
        self.tree = tree = str2xml(content)  # noqa: S314

        self.response_description: Optional[str] = prop(tree, "responsedescription")

        self.responses: Dict[str, Response] = {}
        for resp in tree.findall(".//{DAV:}response"):
            r_obj = Response(resp)
            self.responses[r_obj.path_norm] = r_obj

    def get_response_for_path(self, hostname: str, path: str) -> Response:
        """Provides response for the resource with the specific href/path.

        Args:
            hostname: Hostname path.
            path: Propfind response could have multiple responses inside
                for multiple resources (could be recursive based on the `Depth`
                as well). We use `href` to match the proper response for that
                resource.
        """
        return self.responses[join_url_path(hostname, path)]

    def raise_for_status(self) -> None:
        """Raise error if the responses in the multistatus resp. has errors."""
        statuses = {
            resp.href: resp.reason_phrase
            for resp in self.responses.values()
            if resp.reason_phrase
            and resp.status_code
            and 400 <= resp.status_code <= 599
        }
        statuses = dict(sorted(statuses.items()))
        if statuses:
            raise MultiStatusResponseError(statuses)


# TODO: support `allprop`?
def prepare_propfind_request_data(
    name: Optional[str] = None, namespace: Optional[str] = None
) -> Optional[str]:
    """Prepares propfind request data from specified name.

    In this case, when sent to the server, the `<prop> will only contain the
    `name` property
    """
    if not name:
        return None
    name = MAPPING_PROPS.get(name) or name
    root = Element("propfind", xmlns="DAV:")
    SubElement(SubElement(root, "prop"), "{DAV:}" + name, xmlns=namespace or "")
    return xml2string(root, encoding="unicode")


def parse_multistatus_response(
    http_response: "HTTPResponse",
) -> MultiStatusResponse:
    """Parses multistatus response from the given http response."""
    if http_response.status_code != 207:
        raise ValueError("http response is not a multistatus response")
    return MultiStatusResponse(http_response.text)



"""Some functional utilities."""

import time
from functools import wraps
from itertools import repeat
from typing import (
    TYPE_CHECKING,
    Callable,
    Iterable,
    Iterator,
    Optional,
    Type,
    TypeVar,
    Union,
)

if TYPE_CHECKING:
    from typing_extensions import ParamSpec

    P = ParamSpec("P")

_T = TypeVar("_T")


def repeat_func(func: Callable[[], _T], times: Optional[int] = None) -> Iterator[_T]:
    """Repeatedly calls a function multiple times.

    It will call infinite times if not specified, otherwise as many times as
    specified in `times` arg.
    """
    args = (None, times) if times else (None,)
    return (func() for i in repeat(*args))


def retry(
    retries: int,
    errors: Iterable[Type[Exception]],
    timeout: Union[float, Callable[[int], float]] = 0,
    filter_errors: Optional[Callable[[Exception], bool]] = None,
) -> Callable[[Callable[[], _T]], _T]:
    """Retry a given function."""
    _errors = tuple(errors)

    def wrapped_function(func: Callable[[], _T]) -> _T:
        for attempt in range(retries):
            try:
                return func()
            except _errors as exc:
                if not (filter_errors is None or filter_errors(exc)):
                    raise
                if attempt + 1 == retries:  # reraise on the last attempt
                    raise

            value = timeout(attempt) if callable(timeout) else timeout
            # TODO: do we need circuit breaking?
            time.sleep(value)

        raise AssertionError("it should never reach here")  # pragma: no cover

    return wrapped_function


def wrap_fn(
    func: "Callable[P, _T]", *args: "P.args", **kwargs: "P.kwargs"
) -> Callable[[], _T]:
    """Wraps a function and turns into a 0-arity function."""

    @wraps(func)
    def wrapped() -> _T:
        return func(*args, **kwargs)

    return wrapped
